@@ -4,10 +4,17 @@ CLI command that handles working with the AV development environment
 
 # Imports from av
 from avtoolbox.utils.logger import LOGGER
-from avtoolbox.utils.files import search_upwards_for_file
+from avtoolbox.utils.yaml_parser import YAMLParser
+from avtoolbox.utils.files import search_upwards_for_file, read_text
 
-# Docker imports
+# Other imports
 from python_on_whales import docker, DockerClient, exceptions as docker_exceptions
+import yaml
+import pkgutil
+import os
+import tempfile
+import socket
+import pathlib
 
 def _run_env(args):
     """
@@ -56,33 +63,141 @@ def _run_env(args):
     LOGGER.info("Running 'dev env' entrypoint...")
 
     # Check docker-compose is installed
+    LOGGER.debug("Checking if 'docker compose' (V2) is installed...")
     if not docker.compose.is_installed():
         LOGGER.fatal("The command 'docker compose' is not installed. See http://projects.sbel.org/avtoolbox/tutorials/using_the_development_environment.html for more information.")
         return
+    LOGGER.debug("'docker compose' (V2) is installed.")
 
+    # Check that the .avtoolbox.conf file is located in this directory or any parent directories
+    # This file should be at the root of any avtoolbox compatible stack
+    LOGGER.info("Searching for '.avtoolbox.yml'...")
+    conf = search_upwards_for_file('.avtoolbox.yml')
+    if conf is None:
+        LOGGER.fatal("No .avtoolbox.yml file was found in this directory or any parent directories. Make sure you are running this command in an avtoolbox compatible repository.")
+        return
+    root = conf.parent
+    LOGGER.info(f"Found '.avtoolbox.yml' at {conf}.")
+
+    # Read the conf file 
+    # The conf file _can_ contain the following information (none required):
+    # 
+    # project: <str> (required)
+    #   This is the name of the project that the docker containers will be named after
+    #   The main development environment will be called {project}-dev and so on.
+	#
+    # user: <dict> (optional)
+    #   name: <str> (optional)
+    #       The user to create in the container. If not set, will grab the current user.
+    #
+    #   uid: <int> (optional)
+    #       The user id to use. If not set, will grab it from the current user. (NOT USED)
+    #
+    #   gid: <int> (optional)
+    #       The group id to use. If not set, will grab it from the current user. (NOT USED)
+    #
+    # docker: <dict> (optional)
+    #   context: <path> (optional)
+    #     This is the directory that holds any docker specific files (like custom dockerfiles or
+    #     packages/scripts/dependencies that needed to be used when building the docker image).
+    #     The path is relative to the .avtoolbox.yml file.
+    #
+    #   dependences: <list> (optional)
+    #     A list of dependencies that should be installed in the docker container.
+    #
+    # vnc: <dict> (optional)
+    #   novnc_port: <int> (optional)
+    #       The port that novnc is displayed on. Defaults to 8080.
+    #
+    #   vnc_port: <int> (optional)
+    #       The port that vnc is displayed on. Defaults to 5900.
+    #
+    #   password: <str> (optional)
+    #       The password to use for vnc. Defaults to the project name.
+    #
+    # ros: <dict> (optional)
+    #   workspace: <path> (optional)
+    #     This is the directory which holds the ROS 2 workspace code
+    #     The path is relative to the .avtoolbox.yml file.
+    #
+    #   distro: <str> (optional)
+    #       The ros 2 distro to use. Must be compatible with osrf/ros distrobutions
+    #       on DockerHub. Defaults to galactic.
+
+    LOGGER.debug("Parsering '.avtoolbox.yml' file.")
+    conf_parser = YAMLParser(conf)
+
+    def get_attr(*args, default=None):
+        if not conf_parser.contains(*args):
+            LOGGER.debug(f"'{'.'.join(args)}' wasn't found in the .avtoolbox.yml file. Assuming '{'.'.join(args)}' to be '{default}'")
+            ret = default
+        else:
+            ret = conf_parser.get(*args)
+        return ret
+
+    def read_data(*args):
+        path = os.path.realpath(os.path.join(__file__, "..", *args))
+        return read_text(path), path
+
+    project = get_attr("project")
+    if project is None:
+        LOGGER.fatal(f"'project' must be provided in '.avtoolbox.yml'. Wasn't found, cannot continue.")
+        return
+
+    username = get_attr("user", "name", default=os.getlogin())
+    uid = get_attr("user", "uid", default=os.getuid())
+    gid = get_attr("user", "gid", default=os.getgid())
+
+    # Docker
+    dev_dockerfile, dev_dockerfile_path = read_data("docker", "dev", "dev.dockerfile")
+    context = root / get_attr("dev", "context", default="docker")
+    apt_dependencies = ' '.join(get_attr("dev", "apt-dependencies", default=""))
+    pip_requirements = ' '.join(get_attr("dev", "pip-requirements", default=""))
+
+    # ROS
+    workspace = get_attr("ros", "workspace", default="workspace")
+    distro = get_attr("ros", "distro", default="galactic")
+    
+    # VNC
+    novnc_port = get_attr("vnc", "novnc_port", default="8080")
+    vnc_port = get_attr("vnc", "vnc_port", default="5900")
+    vnc_password = get_attr("vnc", "password", default=project)
+    vnc_dockerfile, vnc_dockerfile_path = read_data("docker", "vnc", "vnc.dockerfile")
+
+    # The docker containers are generated from a docker-compose.yml file
+    # Grab the shipped configuration files that come with the toolbox package
+    docker_compose, docker_compose_path = read_data("docker", "docker-compose.yml")
+    docker_compose_fmt = {
+        "PROJECT": project,
+        "ROOT": root,
+        "DEV.DOCKERFILE": dev_dockerfile_path,
+        "VNC.DOCKERFILE": vnc_dockerfile_path,
+        "USERNAME": username,
+        "USER": f"{uid}:{gid}",
+        "CONTEXT": context,
+        "APT-DEPENDENCIES": apt_dependencies,
+        "PIP-REQUIREMENTS": pip_requirements,
+        "NOVNC_PORT": novnc_port,
+        "VNC_PORT": vnc_port,
+        "VNC_PASSWORD": vnc_password,
+        "WORKSPACE": workspace,
+        "ROSDISTRO": distro,
+    }
+
+    for k,v in docker_compose_fmt.items():
+        k = f"${{{k}}}"
+        v = str(v)
+        docker_compose = docker_compose.replace(k, v)
+    docker_compose_yml = yaml.load(docker_compose, Loader=yaml.SafeLoader)
+    
     # If no command is passed, start up the container and attach to it
     cmds = [args.build, args.up, args.down, args.attach] 
     if all(not c for c in cmds):
         args.up = True
         args.attach = True
 
-    # Search for a file called docker-compose.yml in any of the parent directories. This file
-    # is what holds the default configuration for the docker compose package. We'll read this file,
-    # update any of the config (if necessary), then generate a temporary file which will be passed to
-    # the docker compose command
-    compose_file = search_upwards_for_file('docker-compose.yml')
-    if compose_file is None:
-        LOGGER.fatal("No docker-compose.yml configuration was found in this directory or any parent directories. Make sure you are running this command in the AV repository.")
-        return
-
-    if any([a not in docker.compose.config().services.keys() for a in ['dev', 'vnc']]):
-        LOGGER.fatal("The docker-compose.yml configuration does not contain the services 'dev' and/or 'vnc'. Make sure you are running this command in the AV repository.")
-        return
-
     # Complete the arguments
     if not args.dry_run:
-        import socket, os, tempfile, yaml
-
         def _is_port_in_use(port: int) -> bool:
             """Helper function to check if a port is currently in use."""
             with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
@@ -101,8 +216,13 @@ def _run_env(args):
             # Create the temporary file and don't let it delete automatically yet
             tmp = tempfile.NamedTemporaryFile(delete=False)
 
-            # Read the docker compose file and grab it's config
-            config = docker.compose.config(return_json=True)
+            # Write the config to the temporary yaml file (first time)
+            with open(tmp.name, 'w') as yaml_file:
+                yaml.dump(docker_compose_yml, yaml_file)
+
+            # Set the compose file to the temporary file
+            client = DockerClient(compose_files=[tmp.name])
+            config = client.compose.config(return_json=True)
 
             # Get main dev name
             if not any('dev' in service for service in config['services']):
@@ -114,7 +234,17 @@ def _run_env(args):
                     dev_name = service['container_name']
                     break
 
+            if args.down:
+                LOGGER.info(f"Tearing down...")
+                client.compose.down()
+
+            if args.build:
+                LOGGER.info(f"Building...")
+                client.compose.build()
+
             if args.up:
+                LOGGER.info(f"Spinning up...")
+
                 # Ignore all running services
                 config['services'] = {name : service for name, service in config['services'].items() if len(docker.container.list(filters={"name": service["container_name"]})) == 0 }
 
@@ -141,22 +271,8 @@ def _run_env(args):
                                 port['published'] += 1
                             break
 
-            # Write the config to the temporary yaml file
-            with open(tmp.name, 'w') as yaml_file:
-                yaml.dump(config, yaml_file)
-
-            # Now set the compose file to the temporary file
-            client = DockerClient(compose_files=[tmp.name])
-
-            if args.down:
-                LOGGER.info(f"Tearing down...")
-                client.compose.down()
-            if args.build:
-                LOGGER.info(f"Building...")
-                client.compose.build()
-            if args.up:
-                LOGGER.info(f"Spinning up...")
                 client.compose.up(detach=True)
+
             if args.attach:
                 LOGGER.info(f"Attaching...")
                 try:
